/// @file
////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2016/17      Christian Lessig, Otto-von-Guericke Universitaet Magdeburg
///
////////////////////////////////////////////////////////////////////////////////////////////////////
///
///  module     : lecture 8
///
///  author     : lessig@isg.cs.ovgu.de
///
///  project    : GPU Programming
///
///  description: CUDA basics
///
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <algorithm>

#include "cuda_util.h"
#include <cuda_runtime.h>
#include <cuda.h>

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Non-local means filter in 1D
////////////////////////////////////////////////////////////////////////////////////////////////////
template< int ApronSize >
void
meanRef( float* data, float* means, int n) {

  for( int i = 0; i < n; ++i) {
    means[i] = 0.0;
    for( int k = i-ApronSize; k <= i+ApronSize; ++k) {
      if( k >= 0 && k < n) {
        means[i] += data[k];
      }
    }
    means[i] /= 2*ApronSize+1;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Non-local means filter in 1D
////////////////////////////////////////////////////////////////////////////////////////////////////
template< int ApronSize >
__global__
void
mean( float* gdata, int n) {

  __shared__ float data[1024 + 2*ApronSize];

  // read data to shared memory

  int tid = blockIdx.x * blockDim.x + threadIdx.x;
  data[threadIdx.x + ApronSize] = gdata[tid];

  // read left apron
  if( tid >= ApronSize) {
    if( threadIdx.x < ApronSize) {
      data[threadIdx.x] = gdata[tid - ApronSize];
    }
  }
  else {
    data[threadIdx.x] = 0.0;
  }

  // read right apron
  if( (tid + ApronSize) < n) {
    if( threadIdx.x >= blockDim.x-ApronSize) {
      int ctid = ApronSize - (blockDim.x - threadIdx.x);
      data[blockDim.x + ApronSize + ctid] = gdata[tid + ApronSize];
    }
  }
  else {
    int ctid = ApronSize - (blockDim.x - threadIdx.x);
    data[blockDim.x + ApronSize + ctid] = 0.0;
  }

  // since we share data all threads have to finish reading before we can proceed
  __syncthreads();

  // compute mean

  float sum = 0.0;
  for( int i = ((int) threadIdx.x) - ApronSize; i < ((int) threadIdx.x) + ApronSize + 1; ++i) {
    sum += data[i + ApronSize];
  }
  sum /= static_cast<float>(2*ApronSize+1);

  gdata[tid] = sum;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// program entry point
////////////////////////////////////////////////////////////////////////////////////////////////////
int
main( int /*argc*/, char** /*argv*/ ) {

  const int n = 32;
  float* data = (float*) malloc( sizeof(float) * n);
  for( unsigned int i = 0; i < n; ++i) {
    data[i] = (float) i;
  }

  // compute reference solution
  float* meansref = (float*) malloc( sizeof(float) * n);
  meanRef<2>( data, meansref, n);

  // check execution environment
  int deviceCount = 0;
  checkErrorsCuda( cudaGetDeviceCount(&deviceCount));
  if( 0 == deviceCount) {
    std::cerr << "initDevice() : No CUDA device found." << std::endl;
    return EXIT_FAILURE;
  }

  // query the device properties
  cudaDeviceProp devProp;
  cudaGetDeviceProperties(&devProp, 0);
  printDeviceProps( devProp);

  // set the device
  int device_handle = 0;
  cudaSetDevice( device_handle);

  // allocate device memory
  float* data_device = nullptr;
  checkErrorsCuda( cudaMalloc((void **) &data_device, sizeof(float) * n));

  // copy device memory
  checkErrorsCuda( cudaMemcpy( (void*) data_device, data, sizeof(float) * n, cudaMemcpyHostToDevice ));

  // determine thread layout
  const int MAX_THREADS_PER_BLOCK = 1024;
  int num_threads_per_block = std::min( n, MAX_THREADS_PER_BLOCK);
  int num_blocks = n / MAX_THREADS_PER_BLOCK;
  if( 0 != n % MAX_THREADS_PER_BLOCK) {
    num_blocks++;
  }
  std::cout << "num_blocks = " << num_blocks << " :: "
            << "num_threads_per_block = " << num_threads_per_block << std::endl;

  // run kernel
  mean<2><<< num_blocks , num_threads_per_block >>>( data_device, n);
  cudaDeviceSynchronize();

  checkLastCudaError( "Kernel launch failed.");

  // copy result back to host
  checkErrorsCuda( cudaMemcpy( data, data_device, sizeof(float) * n, cudaMemcpyDeviceToHost ));

  // output
  bool correct = true;
  for( unsigned int i = 0; i < n; ++i) {
    correct &= (data[i] == meansref[i]);
  }
  std::cout << (correct ? "correct" : "incorrect") << std::endl;

  // clean up device memory
  checkErrorsCuda( cudaFree( data_device));

  // clean up host memory
  free( meansref);
  free( data);

  return EXIT_SUCCESS;
}
